#ifndef BOARD_CLASS
#define BOARD_CLASS

#include "square.h"

#define COLS 7
#define ROWS 6

class Board
{
private:
	Board();
	
	static Board *instance;
	int colFill[7];
	int moves[42];
	int totalMoves;
	int totalCommited;
	
public:
	Square squares[7][7];
	
	Square *cols[21][4];
	
	Square *rows[24][4];
	Square *lowerRows[24][4];
	
	Square *diags[24][4];
	Square *lowerDiags[24][4];
	
	static inline Board *getInstance()
	{
		if(instance==NULL) instance=new Board;
		
		return instance;
	}
	
	void display();
	
	inline bool isSymmetrical()
	//Returns true if the board has vertical symmetry and thus we can search half the board for extra speed
	{
		for(int iter1=0; iter1<COLS/2; iter1++)
		{
			for(int iter2=0; iter2<ROWS; iter2++)
			{
				if(squares[iter1][iter2]!=squares[COLS-iter1-1][iter2]) return false;
			}
		}
		
		return true;
	}
	
	inline bool move(int col, Square color)
	//Returns false if column is full or if color==sqEmpty || color==sqNull (though why the hell you would pass sqNull or sqEmpty is beyond me...)
	{
		if(color==sqEmpty || color==sqNull) return false;
		
		if(colFill[col]<0) return false;
		
		squares[col][colFill[col]]=color, 
		moves[totalMoves]=col;
		
		colFill[col]--;
		totalMoves++;
		
		return true;
	}
	
	inline bool undo()
	//Returns false if no moves are left to undo
	{
		if(totalMoves==0 || totalMoves<=totalCommited) return false;
		
		totalMoves--;
		colFill[moves[totalMoves]]++;
		
		squares[moves[totalMoves]][colFill[moves[totalMoves]]]=sqEmpty;
		
		return true;
	}
	void commit();	//Specifies that all moves are fixed and can't be undone
	
	inline Square victory()
	//Returns color of victor or sqNull if no victor yet or sqEmpty if its a tie
	{
		int iter;
		
		Board *board;
		
		board=Board::getInstance();
		
		for(iter=0; iter<21; iter++)
		{
			if(*board->cols[iter][0]!=sqEmpty && *board->cols[iter][0]==*board->cols[iter][1] && *board->cols[iter][0]==*board->cols[iter][2] && *board->cols[iter][0]==*board->cols[iter][3]) return *board->cols[iter][0];
		}
		
		for(iter=0; iter<24; iter++)
		{
			if(*board->rows[iter][0]!=sqEmpty && *board->rows[iter][0]==*board->rows[iter][1] && *board->rows[iter][0]==*board->rows[iter][2] && *board->rows[iter][0]==*board->rows[iter][3]) return *board->rows[iter][0];
		}
		
		for(iter=0; iter<24; iter++)
		{
			if(*board->diags[iter][0]!=sqEmpty && *board->diags[iter][0]==*board->diags[iter][1] && *board->diags[iter][0]==*board->diags[iter][2] && *board->diags[iter][0]==*board->diags[iter][3]) return *board->diags[iter][0];
		}
		
		//Check for no tie
		for(iter=0; iter<COLS; iter++)
		{
			if(colFill[iter]>=0) return sqNull;
		}
		
		//Every square is filled; we have a tie
		return sqEmpty;
	}	
};

#endif
