#include <stdlib.h>
#include <iostream.h>
#include <math.h>

#include "debug.h"
#include "break.h"
#include "board.h"
#include "square.h"
#include "heuristic.h"
#include "bd_string.h"

int searchOrder[7];

HeurFunc promptForHeuristic()
{
	String inputLine;
	
	cout << "0: Dr. Moore" << BREAK << endl;
	cout << "1: Bob, the Wonder Llama" << BREAK << endl;
	cout << "2: A lumberjack" << BREAK << endl;
	cout << "3: Tux, the Penguin" << BREAK << endl;
	cout << "4: David" << BREAK << endl;
	cout << "5: Mr. Anderson" << BREAK << endl;
	cout << "6: Luthien" << BREAK << endl;
	cout << "7: Beren" << BREAK << endl;
	inputLine.getLine(cin, 1024);
	
	while(!inputLine.isNumber() || inputLine.toLong()<0 || inputLine.toLong()>7)
	{
		cout << "\aThat is not a valid menu option!" << BREAK << endl;
		cout << "Select a player:" << BREAK << endl;
		
		cout << "0: Dr. Moore" << BREAK << endl;
		cout << "1: Bob, the Wonder Llama" << BREAK << endl;
		cout << "2: A lumberjack" << BREAK << endl;
		cout << "3: Tux, the Penguin" << BREAK << endl;
		cout << "4: David" << BREAK << endl;
		cout << "5: Mr. Anderson" << BREAK << endl;
		cout << "6: Luthien" << BREAK << endl;
		cout << "7: Beren" << BREAK << endl;
		
		inputLine.getLine(cin, 1024);
	}
	
	switch(inputLine.toLong())
	{
		case 0:
			return mooreHeur;
			
		case 1:
			return testHeur1;
			
		case 2:
			return testHeur2;
			
		case 3:
			return testHeur3;
			
		case 4:
			return testHeur4;
			
		case 5:
			return testHeur5;
			
		case 6:
			return testHeur6;
			
		case 7:
			return testHeur7;
	}
	
	//Theoretically, this should never get executed:
	return NULL;
}

String getHeuristicName(HeurFunc heur)
{
	if(heur==mooreHeur) return "Dr. Moore";
	if(heur==testHeur1) return "Bob, the Wonder Llama";
	if(heur==testHeur2) return "A lumberjack";
	if(heur==testHeur3) return "Tux, the Penguin";
	if(heur==testHeur4) return "David";
	if(heur==testHeur5) return "Mr. Anderson";
	if(heur==testHeur6) return "Luthien";
	if(heur==testHeur7) return "Beren";
	
	//Theoretically, this should never get executed:
	return "";
}

int guessBestCol(Square color, int depth, HeurFunc heurFunc)
{
	int bestCol[7];
	int numBestCols;
	int bestValue;
	int currValue;
	int iterInc;
	bool foundMove;
	bool checkMate;
	Board *board;
	int iter;
	int iterCol;
	
	board=Board::getInstance();
	
	if(board->victory()!=sqNull || color==sqNull || color==sqEmpty) return -1;
	
	numBestCols=0;
	foundMove=false;
	
	if(board->isSymmetrical()) iterInc=1;
	else iterInc=1;
	
	checkMate=false;
	
	do
	{
		//Use iterCol so that searching starts in column 3. This happens to make alpha-beta pruning more effective.
		for(iter=0, iterCol=searchOrder[0]; iter<COLS; iter+=iterInc, iterCol=searchOrder[iter])
		{
			if(board->move(iterCol, color))
			{
				if(board->victory()!=sqNull)
				{
					board->undo();
					
					return iterCol;	//Either its our win immediately so take it or it leads to a tie in which case there is no other choise but this one
				}
				
				currValue=getMiniMaxValue((color==sqWhite)?sqBlack:sqWhite, depth-1, heurFunc, -HEURMAX, HEURMAX);
				
				if(!foundMove || (color==sqBlack && currValue<bestValue) || (color==sqWhite && currValue>bestValue))
				{
					foundMove=true;
					
					bestCol[0]=iterCol;
					numBestCols=1;
					
					bestValue=currValue;
				}
				else if(currValue==bestValue)
				{
					bestCol[numBestCols]=iterCol;
					numBestCols++;
				}
				
				board->undo();
			}
		}
		
		if(checkMate) depth++;
		
		if(bestValue==((color==sqWhite)?HEURMAX:-HEURMAX) && depth && !checkMate)
		{
			checkMate=true;
			
			bestValue=0;
			depth=0;
		}
	}
	//Loop if we have an insta-win; search for shortest deoth to kill so that we do it in fewest turns possible.
	while(checkMate && bestValue!=((color==sqWhite)?HEURMAX:-HEURMAX));
	
	if(checkMate && depth) cout << "You are doomed in " << (depth+1)/2 << "..." << BREAK << endl;
	
	//If symmerical, we searched only half the board. Randomly decide which side to move on.
	if(board->isSymmetrical() && rand()%2) bestValue=7-bestValue;
	
	if(bestValue==((color==sqWhite)?-HEURMAX:HEURMAX) && depth)
	{
#ifdef DEBUG
		cout << "DEBUG: Looks hopeless; trying to delay death." << BREAK << endl;
#endif
		
		return guessBestCol(color, depth-1, heurFunc);
	}
	
#ifdef DEBUG
	cout << "DEBUG: Best guess heuristic: " << bestValue << "." << BREAK << endl;
	cout << "DEBUG: Num of best moves: " << numBestCols*((board->isSymmetrical())?2:1)-((board->isSymmetrical() && bestCol[0]==3)?1:0) << "." << BREAK << endl;
#endif
	
	return bestCol[rand()%numBestCols];
}
