#include <stdlib.h>
#include <time.h>
#include <iostream.h>
#include <math.h>

#include "debug.h"
#include "break.h"
#include "board.h"
#include "square.h"
#include "heuristic.h"
#include "bd_string.h"

int main(int args, char *argv[])
{
	String inputLine;		//Used solely for nabbing input from the user because "cin >>" sux ass
	int turn;				//Running total of number of turns the game has gone
	int humans;				//The number of human players in the game (0-2)
	
	int currentPlayer;		//0==human; 1==computer (when one human player)
	Square colors[2];		//colors[i]==color of player i (see currentPlayer comment)
	int levels[2];			//levels[i]==skill level for the player i (if its computer player)
	HeurFunc heurs[2];		//heurs[i]==heuristic function to use the player i (if its computer player)
	int timeMod[2];			//timeMod[i]==time modifier for player i (if its a computer player) - when the computer sees it's finding moves very quickly as if often the case in late game, it will increase the search depth
	time_t timeLimit[2];		//timeMod[i]==time limit for player i (if its a computer player) - when we start getting faster than this then we start bumping up timeMod (initially set to however long the first full depth search moves take)
	
	time_t currTime;			//Holds how long the last turn took the current computer player; used to figure out when to change timeMod[i]
	
 	int emptyCols;			//Will store how many columns on the board have been filled up
	int actualDepth;		//Stores the actual depth the computer will search; this starts at what the users specified and increases as columns get filled up since when a column is filled the number of states to search decreases
	int bestColForComputer;	//Will hold the column the computer chose to move in for the turn
	
	int iter;				//Used in for loops
	
	bool pauseAfterComp;
	
	Board *board;			//The actual Connect 4 board
	
	//Seed the randomizer
	srand(time(NULL));
	
	board=Board::getInstance();
	
	//Init the search order for the board. This order seems best for alpha-beta pruning to work really well.
	//Basically goes from the center out: 3 4 2 5 1 6 0.
	for(iter=0; iter<COLS; iter++) searchOrder[iter]=((iter%2)*2-1)*(iter+1)/2+3;
	
	timeMod[0]=0;
	timeMod[1]=0;
	
	pauseAfterComp=false;
	
	cout << BREAK << endl;
	cout << " ###    ###   #   #  #   #  #####   ###   #####        4" << BREAK << endl;
	cout << "#   #  #   #  ##  #  ##  #  #      #   #    #       4  4" << BREAK << endl;
	cout << "#      #   #  # # #  # # #  ####   #        #       44444" << BREAK << endl;
	cout << "#   #  #   #  #  ##  #  ##  #      #   #    #          4" << BREAK << endl;
	cout << " ###    ###   #   #  #   #  #####   ###     #          4" << BREAK << endl;
	cout << BREAK << endl;
	cout << "(c) 2002 BoneHead Productions." << BREAK << endl;
	cout << BREAK << endl;
	
	cout << "Enter the number of human players (0-2)." << BREAK << endl;
	
	inputLine.getLine(cin, 1024);
	
	while(!inputLine.isNumber() || inputLine.toLong()<0 || inputLine.toLong()>2)
	{
		cout << "\aThat is not a valid number of computer players." << BREAK << endl;
		cout << "Enter the number of human players (0-2)." << BREAK << endl;
		
		inputLine.getLine(cin, 1024);
	}
	
	humans=inputLine.toLong();
	
	switch(humans)
	{
		case 2:
		{
			humans=2;
			
			currentPlayer=0;
			colors[0]=sqBlack;
			colors[1]=sqWhite;
			
			break;
		}
		case 1:
		{
			cout << "Select a person for for the computer player:" << BREAK << endl;
			heurs[1]=promptForHeuristic();
			
			cout << "Enter a skill level from 1 or higher." << BREAK << endl;
			inputLine.getLine(cin, 1024);
			
			while(!inputLine.isNumber() || inputLine.toLong()<1)
			{
				cout << "\aThat is not a valid skill level!" << BREAK << endl;
				cout << "Enter a skill level from 1 or higher." << BREAK << endl;
				
				inputLine.getLine(cin, 1024);
			}
			
			levels[1]=inputLine.toLong();
			
			cout << "Type \"me\" if you are going first." << BREAK << "\nOtherwise enter anything else." << BREAK << endl;
			
			if(inputLine.getLine(cin, 1024).stripLeading().stripTrailing().toLower()=="me")
			{
				currentPlayer=0;
				
				colors[0]=sqBlack;
				colors[1]=sqWhite;
			}
			else
			{
				currentPlayer=1;
				
				colors[0]=sqWhite;
				colors[1]=sqBlack;
			}
			
			break;
		}
		case 0:
		{
			currentPlayer=0;
			
			colors[0]=sqBlack;
			colors[1]=sqWhite;
			
			cout << "Select a person for the first computer player:" << BREAK << endl;
			heurs[0]=promptForHeuristic();
			
			cout << "Enter a skill level from 0 or higher for the first player." << BREAK << endl;
			inputLine.getLine(cin, 1024);
			
			while(!inputLine.isNumber() || inputLine.toLong()<0)
			{
				cout << "\aThat is not a valid skill level!" << BREAK << endl;
				cout << "Enter a skill level from 0 or higher for the first player." << BREAK << endl;
				
				inputLine.getLine(cin, 1024);
			}
			
			levels[0]=inputLine.toLong();
			
			cout << "Select a Select a person for the second computer player:" << BREAK << endl;
			heurs[1]=promptForHeuristic();
			
			cout << "Enter a skill level from 0 or higher for the second player." << BREAK << endl;
			inputLine.getLine(cin, 1024);
			
			while(!inputLine.isNumber() || inputLine.toLong()<0)
			{
				cout << "\aThat is not a valid skill level!" << BREAK << endl;
				cout << "Enter a skill level from 0 or higher for the second player." << BREAK << endl;
				
				inputLine.getLine(cin, 1024);
			}
			
			levels[1]=inputLine.toLong();
			
			cout << "Enter \"no\" if you do not want a pause between moves." << BREAK << endl;
			
			if(inputLine.getLine(cin, 1024).stripLeading().stripTrailing().toLower()!="no") pauseAfterComp=true;
			
			break;
		}
	}
	
	turn=0;
	
	while(board->victory()==sqNull)
	{
		if((currentPlayer==0 && humans==1) || humans==2)
		//This player is the human or the computer player who went first if there is no human
		{
			board->display();
			cout << "It is your move." << BREAK << endl;
			
			do
			{
				if(colors[currentPlayer]==sqWhite) cout << "You are O." << BREAK << endl;
				else cout << "You are *." << BREAK << endl;
				
				cout << "Enter a column:" << BREAK << endl;
				
				inputLine.getLine(cin, 1024);
				
				if(!inputLine.isNumber() || inputLine.toLong()<0 || inputLine.toLong()>=COLS)
				{
					cout << "\aEnter a number from 0 to " << COLS-1 << "!" << BREAK << endl;
					board->display();
				}
				else
				{
					if(board->move(inputLine.toLong(), colors[currentPlayer])) board->commit();
					else
					{
						cout << "\aThat column is full. Pick another." << BREAK << endl;
						board->display();
						
						inputLine="xxx";
					}
				}
			}
			while(!inputLine.isNumber() || inputLine.toLong()<0 || inputLine.toLong()>=COLS);
			
			cout << BREAK << endl;
		}
		else
		//This handles the turn of the computer players
		{
			for(emptyCols=7, iter=0; iter<COLS; emptyCols=((board->squares[iter][0]==sqEmpty)?emptyCols:emptyCols-1), iter++);	//Yes, this is obfuscated, but I'm the only one who is working with this code and I want to have some fun with it.  :)
			if(emptyCols) for(actualDepth=levels[currentPlayer], iter=0; pow(emptyCols, actualDepth)<=pow(7, levels[currentPlayer]) && iter<42; actualDepth++, iter++);	//And I dare you to figure out what this does. Muhahaha!!!
#ifdef DEBUG
			else cout << "\aIt seems to think there are no open columns for the look ahead. This shouldn't happen!" << BREAK << endl;
#endif
			actualDepth--;
			
			if(actualDepth<0) actualDepth=0;
			
			currTime=time(NULL);
			
#ifdef DEBUG
			cout << "DEBUG: Next actual search depth is " << actualDepth << " plus timeMod of " << timeMod[currentPlayer] << ", total " << actualDepth+timeMod[currentPlayer] <<  BREAK << endl;
#endif
			
			bestColForComputer=guessBestCol(colors[currentPlayer], actualDepth+timeMod[currentPlayer], heurs[currentPlayer]);
			
			//See if we are getting much faster
			currTime=time(NULL)-currTime;
			
			if(turn<3) timeLimit[currentPlayer]=currTime;	//Turns 1 and 2 are the first full depth search turns so mark the time it took
			else
			//Else see if we are past the initial 2 turns and are getting much faster then let's boost timeMod to increase search depth for the next time around
			{
				if(currTime<timeLimit[currentPlayer]/2)
				{
					timeMod[currentPlayer]++;
					
#ifdef DEBUG
					cout << "DEBUG: Boosted timeMod to " << timeMod[currentPlayer] << BREAK << endl;
#endif
				}
			}
			
			if(bestColForComputer<0)
			{
				cout << "\aThe computer failed to pick a column." << BREAK << endl;
				
				return 1;
			}
			
			if(!board->move(bestColForComputer, colors[currentPlayer]))
			{
				cout << "\aThere was an error making the computer move." << BREAK << endl;
				
				return 1;
			}
			
			board->commit();
			
			cout << getHeuristicName(heurs[currentPlayer]) << " chose column " << bestColForComputer << "." << BREAK << '\n' << BREAK << endl;
			
			if(!humans)
			{
				board->display();
				
				if(pauseAfterComp)
				{
					cout << "Paused..." << BREAK << endl;
					inputLine.getLine(cin, 1024);
				}
			}
 		}
		
		currentPlayer=(currentPlayer+1)%2;
		
		turn++;
	}
	
	cout << BREAK << "\n\aThe game is over. Final board layout:" << BREAK << endl;
	
	board->display();
	
	if(board->victory()==sqEmpty) cout << "The game is a tie." << BREAK << endl;
	else if(board->victory()==colors[0])
	{
		if(humans==2) cout << "* wins!" << BREAK << endl;	//colors[0]==sqBlack always for 2 human players
		else if(humans==1) cout << "You win!" << BREAK << endl;
		else cout << getHeuristicName(heurs[0]) << " wins as player 1." << BREAK << endl;
	}
	else 
	{
		if(humans==2) cout << "O wins!" << BREAK << endl;	//colors[1]==sqWhite always for 2 human players
		else if(humans==1) cout << "The computer wins." << BREAK << endl;
		else cout << getHeuristicName(heurs[1]) << " wins as player 2." << BREAK << endl;
	}
	
	cout << "The game lasted " << turn << " turns." << BREAK << endl;
	
	cout << "Press RETURN to end." << BREAK << endl;
	inputLine.getLine(cin, 1024);
	
	return 0;
}
